/****************************************************************************
	File: EgBase.cpp
	Created: 8/7/2001
	Modified: 10/26/2001
	
	Description:	Interface methods to the Eyegaze System.
						The EgBase object represents the tracker.
	
	Author: Tim Halverson, Erik Brown
	Principle Investigator: Anthony Hornof
	Cognitive Modeling and Eye Tracking Lab
	CIS Department
	University of Oregon
	
	This is a modification of:
		EgWin.C       Windows interface functions to the Eyegaze System
		Created 11/11/1998, LC Technologies, Inc.

	Modifications: see EgBase.h
		
	Copyright (c) 2007 by the University of Oregon.
	ALL RIGHTS RESERVED.

	Permission to use, copy, and distribute this software in
	its entirety for non-commercial purposes and without fee,
	is hereby granted, provided that the above copyright notice
	and this permission notice appear in all copies and their
	documentation.

	Software developers, consultants, or anyone else who wishes
	to use all or part of the software or its documentation for
	commercial purposes should contact the Technology Transfer
	Office at the University of Oregon to arrange a commercial
	license agreement.

	This software is provided "as is" without expressed or
	implied warranty of any kind.
*****************************************************************************/

#include "EgBase.h"
#include "ext.h"

// Size of the container. What is a good number to use? If client reads the
// data too slowly, then the buffer will fill up even with hundreds of
// available slots. If the client reads the data too quickly, we'll probably
// never need more than two slots (maybe just one). If the container does fill
// up, and we have a large container, then we'll be retrieving old data. Seems
// best to keep the container small and "lose" old data and be assured of 
// getting pretty recent data.
const int EgBase::numDataPoints = 10;

EgBase::EgBase()
{
	currentState = STATE_START;
		
	return;
}

EgBase::~EgBase()
{
}

bool EgBase::isConnected()
{
	return (currentState == STATE_CONNECTED || currentState == STATE_RECEIVING);
}

EgStatus EgBase::connect(const char *commName)
{
	EgStatus err = EG_NOERROR;

	if (! isConnected())
	{
		///////////////////////////////////////////
		//
		// Create the socket to the eye tracker.
		//
		struct sockaddr_in	sa;
		struct hostent			*hp;

		// IP address or name was passed in.
		if ( (hp = gethostbyname(commName)) != NULL)
		{
			// Set fields in sockaddr_in struct including the port
			// we connect through.
			memset(&sa, 0, sizeof(sa));
			memcpy((char*)&sa.sin_addr, hp->h_addr, hp->h_length);
			sa.sin_family = hp->h_addrtype;
			sa.sin_port   = htons((u_short)3999);
			
			// Create the socket.
			if ( (theSocket = socket(hp->h_addrtype, SOCK_STREAM, 0)) >= 0)
			{
				// And connect.
				if (::connect(theSocket, (struct sockaddr*)&sa, sizeof sa) != 0)
				{
					printf("Could not connect to the eye tracker.\n");
					
					this->closeSocket();

					err = EG_ERROR;
				}
			}
			else
			{
				printf("Could not create the socket.\n");
				err = EG_ERROR;
			}
		}
		//
		/////////////////////////////////////////////////////////////////////////////////////


		//////////////////////////////////////////////////////////////////////////
		//
		// Now that we have a connection, start the thread that will read data
		// sent by the eye gaze server.
		if (err == EG_NOERROR)
		{
			startThread();
			currentState = STATE_CONNECTED;
		}
		//
		//////////////////////////////////////////////////////////////////////////
	}

	return err;
}

EgStatus EgBase::disconnect()
{
	if (isConnected())
	{
		stopData();
		sendMessage(EG_MESSAGE_TYPE_CLOSE_AND_RECYCLE, NULL, 0);

		// <!-- Windows hack -->
		// Wait a bit to close the socket else we don't seem to get
		// the CLOSE_AND_RECYCLE message to the eye tracker. Mac side
		// works fine without this, so I'm not exactly sure what the 
		// problem is, but doing this won't hurt the Mac side.
		this->Sleep(250);

		closeSocket();
	}

	currentState = STATE_START;

	return EG_NOERROR;
}


EgStatus EgBase::startData()
{
	EgStatus err = EG_RECEIVEFAILED;
	
	if (isConnected())
	{
		// Tell eye tracker to start sending data.
		sendMessage(EG_MESSAGE_TYPE_BEGIN_SENDING_DATA, NULL, 0);
		err = EG_NOERROR;
	}
	
	return err;
}

EgStatus EgBase::stopData()
{
	// Tell server to stop sending data.
	sendMessage(EG_MESSAGE_TYPE_STOP_SENDING_DATA, NULL, 0);
	
	// Then clear the data deque.
	WaitMutex();
	egDataDeque.clear();
	SignalMutex();
	
	return EG_NOERROR;
}

/////////////////////////////////////////////////////////////////////////////////////
//
// The owner of the EgBase object will probably be calling this at regular intervals.
//
bool EgBase::GetData(stEgData &data)
{
	bool ret = false;
	
	// Get exclusive access from where we're reading the data. Remember, two threads
	// are accessing this area, one reading and one writing.
	WaitMutex();
	
	if (egDataDeque.size() > 0)
	{
		data = egDataDeque.front();
		
		egDataDeque.pop_front();
		
		ret = true;
	}
	
	// Make sure to release the mutex or nothing else will get access.
	SignalMutex();
	
	return ret;
}

void EgBase::SetData(stEgData &data)
{
   // Get exclusive access from where we're reading the data. Remember, two threads
	// are accessing this area, one reading and one writing.
	WaitMutex();
   
	egDataDeque.push_back(data);
	
   // Make sure to release the mutex or nothing else will get access.
	SignalMutex();
	
	return;
}

void EgBase::sendMessage(int iMessage, unsigned char *pszBuffer, unsigned int uiBufferLen)
{
	unsigned char auchMessage[2000];
	unsigned char checksum;
	unsigned int  uiMessageLen;
	unsigned char auchMessageLen[3];
	unsigned int i;
	
	uiMessageLen = uiBufferLen + 5;
	auchMessageLen[0] = (uiMessageLen & 0x00FF0000) >> 16;
	auchMessageLen[1] = (uiMessageLen & 0x0000FF00) >> 8;
	auchMessageLen[2] = (uiMessageLen & 0x000000FF);
	
	// Accumulate the checksum of the message header.
	checksum = auchMessageLen[0] +
		auchMessageLen[1] +
		auchMessageLen[2] +
		iMessage;
	
	// Create the header portion of the output buffer (message).
	sprintf((char*)auchMessage,"%c%c%c%c",auchMessageLen[0],
			  auchMessageLen[1],
			  auchMessageLen[2],
			  iMessage);
	
	// Accumulate the checksum over the entire buffer and create the rest of the message.
	for (i=0; i<uiBufferLen; i++)
	{
		checksum += pszBuffer[i];
		auchMessage[i+4] = pszBuffer[i];
	}
	
	// Tack the checksum on at the end of the message.
	auchMessage[i+4] = checksum;
	
	writeSocket((const char*)auchMessage, 5 + uiBufferLen);
}

/////////////////////////////////////////////////////////////////////////////////////
//
// All the thread does is to read from the given socket, and when data is available
// on the socket, it translates the message and stores it in the EgMac object.
//
// Currently, the only messages it traps are GAZEINFO messages -- all others are ignored.
//
#ifdef WIN_VERSION
	DWORD WINAPI EyeTrackerReceive(void* td)
#elif MAC_VERSION
	void* EyeTrackerReceive(void* td)
#endif
{
	EgBase *egBase	= reinterpret_cast<EgBase*>(td);
	
	int bytesRead = -1;
	char message[4096];
	char messageLengthArray[3];
	
	while(true)
	{
		// Prepare for next message by zeroing out these.
		memset(message, '\0', sizeof(message));
		memset(messageLengthArray, '\0', sizeof(messageLengthArray));
		
		// Messages from the eye tracker will first specify the length of the
		// message in the first three bytes. Read one byte at a time. We can 
		// block because we're in a thread devoted to reading from the socket.
		if ( (bytesRead = egBase->readSocket(&messageLengthArray[0], 1)) <= 0 )
			break;
		
		if ( (bytesRead = egBase->readSocket(&messageLengthArray[1], 1)) <= 0 )
			break;
		
		if ( (bytesRead = egBase->readSocket(&messageLengthArray[2], 1)) <= 0 )
			break;
		
		// Now calculate the message length and then read that many bytes
		// from the socket.
		long messageLength = ((unsigned int)messageLengthArray[0] * 65536) +
			((unsigned int)messageLengthArray[1] *   256) +
			((unsigned int)messageLengthArray[2]);
		
		if ( (bytesRead = egBase->readSocket(&message[0], messageLength)) <= 0 )
			break;
		

		// The first byte will specify the message type.
		switch (message[0])
		{
			case EG_MESSAGE_TYPE_GAZEINFO:
				// Starting from the end, get a piece of info, then zero out that position.
				// This way, we can use sscanf() to read in information contained lower in
				// the message array.
				
				
				stEgData egData;
				memset(&egData, '\0', sizeof (stEgData));
				
				// Throw away the checksum.
				message[15] = '\0';
				
				
				// Get delta camera field count.
				int deltaCameraFieldCount;
				sscanf( (char*)&message[13], "%i", &deltaCameraFieldCount );
				
				message[13] = '\0';
				
				
				// Get gaze found boolean.
				int gazeFound;
				sscanf( (char*)&message[12], "%i", &gazeFound );
				
				egData.bGazeVectorFound = (gazeFound) ? true : false;
				message[12] = '\0';
				
				
				// Get pupil radius.
				int pupilRadius;
				sscanf( (char*)&message[9], "%i", &pupilRadius );
				
				egData.fPupilRadiusMm = pupilRadius / 100.0f;
				message[9] = '\0';
				
				
				// Get eye positions.
				sscanf( (char*)&message[5], "%i", &egData.iJGaze );
				message[5] = '\0';
				
				sscanf( (char*)&message[1], "%i", &egData.iIGaze );
				message[1] = '\0';
				
				
				// Increment camera field count.
				egData.ulCameraFieldCount += deltaCameraFieldCount;
				
				egBase->SetData(egData);
				
				break;
				
			default:
				break;
		}
	}
	
	return 0;
}


