/****************************************************************************
	File: EgStructs.h
	Description: Defines structs for EyeGaze system
	Created: 8/7/2001
	
	Author: Tim Halverson
	Principle Investigator: Anthony Hornof
	Cognitive Modeling and Eye Tracking Lab
	CIS Department
	University of Oregon
	
	Based partly on EgWin.h by LC Technologies.

	Copyright (c) 2001 by the University of Oregon.
	ALL RIGHTS RESERVED.

	Permission to use, copy, and distribute this software in
	its entirety for non-commercial purposes and without fee,
	is hereby granted, provided that the above copyright notice
	and this permission notice appear in all copies and their
	documentation.

	Software developers, consultants, or anyone else who wishes
	to use all or part of the software or its documentation for
	commercial purposes should contact the Technology Transfer
	Office at the University of Oregon to arrange a commercial
	license agreement.

	This software is provided "as is" without expressed or
	implied warranty of any kind.
*****************************************************************************/

#ifndef EGSTRUCTS_H
#define EGSTRUCTS_H


/****************************************************************************/
/* The stEgData structure contains the results of the Eyegaze image         */
/* processing from a given field of video camera data.                      */

struct stEgData
{
   bool   bGazeVectorFound;      /* flag indicating whether a valid         */
                                 /*   glint pupil vector was found by the   */
                                 /*   image processing software             */
                                 /*   (TRUE = 1, FALSE = 0)                 */
   int    iIGaze;                /* integer coordinates of the user         */
   int    iJGaze;                /*   gazepoint referenced to the full      */
                                 /*   computer screen (pixels)              */
                                 /*   0,0 origin at upper left corner       */
                                 /*   iIGaze positive rightward             */
                                 /*   iJGaze positive downward              */
   float  fPupilRadiusMm;        /* actual pupil radius (mm)                */
   float  fXEyeballOffsetMm;     /* offset of the eyeball center from       */
   float  fYEyeballOffsetMm;     /*  the camera axis (mm)                   */
                                 /*  Notes on polarity:                     */
                                 /*  x positive: head moves to user's right */
                                 /*  y positive: head moves up              */
   float  fFocusRangeImageTime;  /* distance from the camera sensor plane   */
                                 /*   to the camera focus plane, at the     */
                                 /*   time the camera captured the image    */
                                 /*   (mm)                                  */
   float  fFocusRangeOffsetMm;   /* range offset between the camera focus   */
                                 /*   plane and the corneal surface of the  */
                                 /*   eye - at image time (mm)              */
                                 /*   A positive offset means the eye is    */
                                 /*   beyond the lens' focus range.         */
   float  fLensExtOffsetMm;      /* distance that the lens extension would  */
                                 /*   have to be changed to bring the eye   */
                                 /*   into clear focus (millimeters)        */
                                 /*   (at image time)                       */
   UInt32 ulCameraFieldCount;    /* number of camera fields, i.e. 60ths of  */
								 /*   a second, that have occurred since    */
                                 /*   the starting reference time (midnight */
                                 /*   January 1, this year)                 */
   double dPentiumTSCMicroSec;   /* Pentium Time Stamp Counter, converted   */
                                 /*   to microseconds.                      */
                                 /*   NOTE: we're using Windows to sleep    */
                                 /*   for a second here in order to get a   */
                                 /*   reading on the processor speed.       */
                                 /*   Windows' clock is accurate to 10      */
                                 /*   milliseconds so we should be accurate */
                                 /*   to within about 1% of true.           */
/****************************************************************************/
};

/****************************************************************************/
/* The stEgControl structure contains control and status variables used     */
/* to setup and control the Eyegaze image processing software, and to       */
/* access the Eyegaze data.                                                 */

struct stEgControl
{
                                  /* INPUTS FROM APPLICATION:               */
   stEgData *pstEgData;   		  /* pointer to the first Eyegaze data      */
                                  /*   structure in the Eyegaze data ring   */
                                  /*   buffer.                              */
                                  /*   The application must set the ring-   */
                                  /*   buffer address in the stEgControl    */
                                  /*   structure before calling EgInit()    */
   int    iNDataSetsInRingBuffer; /* number of samples in the application's */
                                  /*   ring buffer of Eyegaze data samples. */
                                  /*   The application must set the ring-   */
                                  /*   buffer length in the stEgControl     */
                                  /*   structure before calling EgInit()    */
   int    iEgBufferTail;          /* buffer index of the last Eyegaze data  */
                                  /*   sample processed by the application  */
                                  /*   The application must update this     */
                                  /*   variable as it uses Eyegaze data.    */
   bool   bTrackingActive;        /* flag controls whether eyetracking is   */
                                  /*   presently on (TRUE = 1) or off       */
                                  /*   (FALSE = 0). If the flag is on when  */
                                  /*   a new camera field finishes, the     */
                                  /*   Eyegaze thread processes the image   */
                                  /*   and puts the results in the data     */
                                  /*   ring buffer; if the flag is off,     */
                                  /*   the camera field is not processed.   */
                                  /*   The application may turn this        */
                                  /*   tracking flag on or off at any       */
                                  /*   time.                                */
   int    iScreenWidthPix;        /* Pixel dimensions of the full           */
   int    iScreenHeightPix;       /*   computer screen                      */
   bool   bEgCameraDisplayActive; /* flag controls whether or not the       */
                                  /*   full 640x480 image from the Eyegaze  */
                                  /*   camera is displayed in a separate    */
                                  /*   window on the VGA.                   */
                                  /*   The application must set this flag   */
                                  /*   prior to calling EgInit() and may    */
                                  /*   turn the display flag on and off     */
                                  /*   at any time.                         */
   int    iEyeImagesScreenPos;    /* Screen position of the eye images.     */
                                  /*   0 = upper left, 1 = upper right      */
   int    iCommType;              /* Communication type:                    */
                                  /*   EG_COMM_TYPE_LOCAL,                  */
                                  /*   EG_COMM_TYPE_SOCKET, or              */
                                  /*   EG_COMM_TYPE_SERIAL.                 */
   char   *pszCommName;           /* Pointer to serial port name or IP      */
                                  /*   address of server machine.           */

                                  /* OUTPUTS TO APPLICATION:                */
   int    iEgBufferHead;          /* buffer index of the most recently      */
                                  /*   processed Eyegaze camera image       */
   bool   bEgBufferOverflow;      /* a flag indicating whether the Eyegaze  */
                                  /*   image processing is more than        */
                                  /*   iNDataSetsInRingBuffer ahead of the  */
                                  /*   application processing the data out  */
                                  /*   of the ring buffer.  Eyegaze sets    */
                                  /*   this flag if iEgBufferHead catches   */
                                  /*   up to iEgBufferTail.*/
	bool bEgBufferUnderflow;	// a flag indicating whether the eyegaze image
								// processing is behind the application processing
								// of data in the ring buffer.  Eyegaze sets this
								// flag if iEgBufferTail catches up with 
								// iEgBufferHead.
   int    iSamplePerSec;          /* Eyegaze image processing rate -        */
                                  /*   depends on the camera field rate:    */
                                  /*   RS_170               60 Hz           */
                                  /*   CCIR                 50 Hz           */
   float  fHorzPixPerMm;          /* Eyegaze monitor scale factors          */
   float  fVertPixPerMm;          /*   (pixel / millimeter)                 */
   void   *pvEgVideoBufferAddress;/* address of the video buffer containing */
                                  /*   the most recently processed camera   */
                                  /*   image field                          */
};

#endif
