///////////////////////////////////////////////////////////////////
//
// File: main.cpp
// 
// Project: eyedata Max external.
// 
// Authors: Erik Brown, Linda Sato.
// Principle Investigator: Anthony Hornof
// Cognitive Modeling and Eye Tracking Lab
// CIS Department, University of Oregon
// 
// Description: Provides the functions that Max calls into. Sets
//              up a thread that reads eye gaze data over a TCP/IP
//              connection with an eye tracker. The external object
//              creates a clock which forces Max to call into the
//              external at regular time intervals and retrieve gaze
//              data.
// 
// Copyright (c) 2007 by the University of Oregon.
// ALL RIGHTS RESERVED.
//
// 
// Permission to use, copy, and distribute this software in
// its entirety for non-commercial purposes and without fee,
// is hereby granted, provided that the above copyright notice
// and this permission notice appear in all copies and their
// documentation.
// 
// Software developers, consultants, or anyone else who wishes
// to use all or part of the software or its documentation for
// commercial purposes should contact the Technology Transfer
// Office at the University of Oregon to arrange a commercial
// license agreement.
// 
// This software is provided "as is" without expressed or
// implied warranty of any kind.
// 
//
///////////////////////////////////////////////////////////////////


#include <string>

#include "ext.h"
#include "ext_strings.h"
#include "commonsyms.h"
#include "ext_obex.h"


/////////////////////////////////////////////////////////////////////////////////////
//
// These #defs are required by Max when writing an external, therefore
// we might as well use them to include the correct .h file for the 
// platform for which we are building.
#ifdef MAC_VERSION
	#include "EgMac.h"
#elif WIN_VERSION
	#include "EgWin.h"
#endif


// This is what our Max/MSP external object looks like.
typedef struct eyedata
{
   t_object    e_ob;
	void        *obex;
   void        *e_out1;       // int outlet for y coordinate      
   void        *e_out2;       // int outlet for x coordinate     
   void        *e_out3;       // float outlet for pupil diameter
   void        *e_out4;       // int outlet for eye found flag   
   void        *e_out5;       // int outlet for line number    
   void        *e_clock;
} t_eyedata;


// Prototypes. Declared as extern "C" because Max says to do so.
extern "C"{
   int      main                 (void);
   void*    eyedata_new          ();
   void     eyedata_free         (t_eyedata *ed);
   void     eyedata_connect      (t_eyedata *ed, t_symbol *ip);
   void     eyedata_disconnect   (t_eyedata *ed);
   void     eyedata_startdata    (t_eyedata *ed);
   void     eyedata_stopdata     (t_eyedata *ed);
   void     eyedata_assist       (t_eyedata *ed, void *b, long msg,long arg, char *dst);
   void     eyedata_retrieve     (t_eyedata *ed);
}


// Globals
t_class  *this_eyedata_object;   // THE Max external object
EgBase   *eg;                    // The communication link with the eye tracker. Thanks to
                                 // inheritance and virtual functions, we can have a base class
                                 // pointer, but have it point to one of the two derived objects,
                                 // either EgMac or EgWin.


std::string eyeTrackerIP = "1.0.0.2";	// default value (pass in ip via 'connect' method)



/////////////////////////////////////////////////////////////////////////////////////
//
// Called by Max when an external is loaded into the Max environment. This function
// basically follows guidelines that all externals follow. Create a new object telling
// Max which are the 'new' and 'free' methods, add methods to the object that tell
// Max when and where to call, register the object, and set the global object pointer
// to the newly created object.
//
int main(void)
{
   t_class *c;

   common_symbols_init();

   c = class_new("eyedata5", (method)eyedata_new, (method)eyedata_free,
                  (short)sizeof(t_eyedata), (method)0L, 0);

   class_obexoffset_set(c, calcoffset(t_eyedata, obex));

   class_addmethod(c, (method)eyedata_connect, "connect", A_DEFSYM, 0L);
   class_addmethod(c, (method)eyedata_disconnect, "disconnect", 0L);
   class_addmethod(c, (method)eyedata_startdata, "startdata", 0L);
   class_addmethod(c, (method)eyedata_stopdata, "stopdata", 0L);
   class_addmethod(c, (method)eyedata_assist, "assist", A_CANT, 0L);

   class_register(CLASS_BOX, c);
   this_eyedata_object = c;
   
   return 0;
}

/////////////////////////////////////////////////////////////////////////////////////
//
// This method is called when our object is created in the Max environment. Allocates
// space for the object and assigns the outlets. Also creates the clock and assigns
// the method to call when the clock interval has elapsed.
//
void* eyedata_new()
{
   t_eyedata *ed = NULL;

   ed = (t_eyedata*)object_alloc(this_eyedata_object);

   if (ed)
   {
      ed->e_out1 = intout(ed);  
      ed->e_out2 = intout(ed);  
      ed->e_out3 = floatout(ed);
      ed->e_out4 = intout(ed);  
      ed->e_out5 = intout(ed); 

      // Create a clock that will call eyedata_retrieve every time it goes off.
      ed->e_clock = clock_new(ed, (method(eyedata_retrieve)));
      
#ifdef MAC_VERSION
      eg = new EgMac;
#elif WIN_VERSION
      eg = new EgWin;
#endif

      post("Eg status: not connected.");
   }

   return (ed);
}


/////////////////////////////////////////////////////////////////////////////////////
//
// Time to get rid of any memory we have allocated using this external object.
//
void eyedata_free(t_eyedata *ed)
{
   if (eg)
   {
      // Make sure to stop the data collection before we get rid of the Eg object.
      eg->disconnect();

      delete eg;
      eg = NULL;
   }
	
   clock_unset(ed->e_clock);
	clock_free((t_object*)ed->e_clock);

   post("object deleted");
}

/////////////////////////////////////////////////////////////////////////////////////
//
// Connect if we haven't done so already. If we're receiving data, then
// this will stop the data!
//
void eyedata_connect(t_eyedata *ed, t_symbol *ip)
{
   EgStatus err = EG_NOERROR;

   // Were we sent the ip address?
   if (strlen(ip->s_name) > 0)
      eyeTrackerIP = ip->s_name; // then override the default

   clock_unset(ed->e_clock); // stop retrieving data

   if (eg)
   {
      err = eg->connect(eyeTrackerIP.c_str());

      post( (err == EG_NOERROR) ? (char*)"Eg status: connected.\n   Choose 'startdata' to begin receiving data" 
                                : (char*)"Eg status: not connected.");
   }
}

/////////////////////////////////////////////////////////////////////////////////////
//
// Disconnect if we are currently connected, otherwise there is no need to 
// do anything.
//
void eyedata_disconnect(t_eyedata *ed)
{
   clock_unset(ed->e_clock);

   EgStatus err = EG_NOERROR;

   if (eg)
      err = eg->disconnect(); // this also stops data collection

   post( (err == EG_NOERROR) ? (char*)"Eg status: not connected." : (char*)"Eg status: failed to disconnect.");
}

/////////////////////////////////////////////////////////////////////////////////////
//
// Start collecting data. We need to be connected to do this, so return a useful
// message if we can't start the data collection.
//
void eyedata_startdata(t_eyedata *ed)
{
   EgStatus err = EG_ERROR;

   if (eg)
   {
      err = eg->startData();

      if (err == EG_NOERROR)
      {
         clock_fdelay(ed->e_clock, 0);
      }
   }

   post( (err == EG_NOERROR) ? (char*)"Eg status: receiving data."
                             : (char*)"Eg status: could not start data.\n   Connect first, then start data.");
}


/////////////////////////////////////////////////////////////////////////////////////
//
// Stop collecting data, but don't disconnect from the eye tracker. Also, we can
// tell Max to stop calling into the external to retrieve data by unsetting the clock.
//
void eyedata_stopdata(t_eyedata *ed)
{
   EgStatus err = EG_ERROR;

   clock_unset(ed->e_clock);

   if (eg)
      err = eg->stopData();

   post( (err == EG_NOERROR) ? (char*)"Eg status: stop receiving data." : (char*)"Eg status: failed to stop data.");
}


/////////////////////////////////////////////////////////////////////////////////////
//
// Status bar, "tooltip" like messages (not sure what Max calls them).
//
void eyedata_assist(t_eyedata *ed, void *b, long msg, long arg, char *dst)
{
   if (msg == ASSIST_INLET)
   {
      strcpy(dst, "accepts connect, disconnect, startdata, and stopdata");
   }
   else if (msg == ASSIST_OUTLET)
   {
      switch (arg)
      {
         case 0:
            strcpy(dst, "int, camera sample no.");
            break;
         case 1:
            strcpy(dst, "int, eye found flag");
            break;
         case 2:
            strcpy(dst, "float, pupil diameter");
            break;
         case 3:
            strcpy(dst, "int, x coord");
            break;
         case 4:
            strcpy(dst, "int, y coord");
            break;
      }
   }
}

/////////////////////////////////////////////////////////////////////////////////////
//
// By setting a clock and an interval, we get Max to call into here. This is where we
// transfer the gaze data collected by this external to Max -- or rather the patcher 
// that uses this external.
//
void eyedata_retrieve(t_eyedata *ed)
{
   if (eg && eg->isConnected())
   {
      stEgData egData;
		
		// Get data while there is data to be gotten.
      while (eg->GetData(egData) == true)
		{
			// ... and pass the data through the outlets.
			outlet_int(ed->e_out1,     egData.iJGaze);
			outlet_int(ed->e_out2,     egData.iIGaze);
			outlet_float(ed->e_out3,   egData.fPupilRadiusMm);
			outlet_int(ed->e_out4,     egData.bGazeVectorFound);
			outlet_int(ed->e_out5,     egData.ulCameraFieldCount);
		}
   }

   // Set the next interval (milliseconds) to get back here.
   //
   // TODO: Determine dynamically how often the EG server sends
   // samples. If we set the delay too low here, we waste CPU, too
   // high, and we create a backlog of samples the EG object must
   // hold onto, not to mention that we will then be retrieving old
   // samples.
   //
	// Update: by experimentation, it doesn't seem to much matter (note very
	// informal benchmarking!).  I've set this value at 8ms, twice as fast as
	// data is collected, and seem to get approximately the same frame rate
	// in Eye Music as when I set this value at 48ms, which is more than
	// twice as slow as data is collected. I'm going to leave this at a value
	// which will get approximately just over one piece of data every time
	// the timer goes off. In other words, for a 16ms collection rate, we
	// going to come in here every 20ms. I can't help but like the idea of
	// polling less often rather than more often.
	//
	// Update 2: This object may be used in conjunction with a fixation object.
	// If that's the case, then in order for the fixations to be accurately
	// determined in real-time, the fixation object needs to get them as soon as
	// possible. This is, of course, due to a design decision. Currently, the 
	// fixation object accepts eye gaze data and assumes this data comes in at
	// known intervals. Since it may be using a velocity-based algorithm, time
	// intervals are a crucial piece of information. A different design, such as
	// passing the time along with a gaze sample, might be better.
   //
	// Let's poll for data at the rate of 120 Hz (8.33 ms intervals).
   clock_fdelay(ed->e_clock, 8.33);
}

