 ///////////////////////////////////////////////////////////////////
 //
 // File: velocity.cpp
 // 
 // Project: fixation Max external.
 // 
 // Authors: Erik Brown.
 // Principle Investigator: Anthony Hornof
 // Cognitive Modeling and Eye Tracking Lab
 // CIS Department, University of Oregon
 // 
 // Description: See velocity.h
 //
 // 
 // Copyright (c) 2007 by the University of Oregon.
 // ALL RIGHTS RESERVED.
 //
 // 
 // Permission to use, copy, and distribute this software in
 // its entirety for non-commercial purposes and without fee,
 // is hereby granted, provided that the above copyright notice
 // and this permission notice appear in all copies and their
 // documentation.
 // 
 // Software developers, consultants, or anyone else who wishes
 // to use all or part of the software or its documentation for
 // commercial purposes should contact the Technology Transfer
 // Office at the University of Oregon to arrange a commercial
 // license agreement.
 // 
 // This software is provided "as is" without expressed or
 // implied warranty of any kind.
 // 
 //
 ///////////////////////////////////////////////////////////////////
 
#include <math.h>
#include "ext.h"
#include "velocity.h"

namespace VF
{
	// Hard coded at 22 inches (558.8mm).
	double Velocity::mmEyeToScreen = 558.8;

	int Velocity::maxIgnoreOutsideFixation = 1;

	
	Velocity::Velocity() :
		threshold(20),
		minFixTime(30),
		sampleRate(60),
		countOutsideFixation(0),
		countIgnoredSamples(0),
		state(VF::MOVING)
	{
		// Need to get screen info, which is accomplished in platform
		// specific ways.
	#ifdef MAC_VERSION
		CGRect r  = CGDisplayBounds(CGMainDisplayID());     // pixels
		CGSize ds = CGDisplayScreenSize(CGMainDisplayID()); // millimeters
		
		pixelsAcrossScreen = r.size.width;
		mmAcrossScreen     = ds.width;
		
	#elif WIN_VERSION
		pixelsAcrossScreen = GetDeviceCaps(GetDC(0), HORZRES);	// pixels
		mmAcrossScreen     = GetDeviceCaps(GetDC(0), HORZSIZE);	// millimeters

	#endif // MAC_VERSION
		
		CalculateNewThreshold();
		
		// What is the "impossible" threshold that we should ignore, in pixels?
		// Use 800 degrees, or 40 * 20 degrees since tan(800) is not
		// nearly the same as 40 * tan(20).
		double mmScreen = 40 * (Velocity::mmEyeToScreen * tan((M_PI * 20) / 180.0));
		
		// Convert to pixels, since that's how we get eye data.
		impossibleThreshold = mmScreen / (mmAcrossScreen / pixelsAcrossScreen);

		return;
	}

	Velocity::~Velocity()
	{
		// We might be in a fixation when deleting -- record it if so.
		if(state == VF::FIX)
		{
			VF::FixationPoint fp;
			VF::Point cg(current_fixation.back());
			
			fp.center_gravity = cg;
			fp.num_points = current_fixation.size();
			
			fixation_points.push_back(fp);
		}
		
		// Do something with our collection of fixation points here.
		// ...

		return;
	}
	
	void Velocity::SetCurrentPoint(VF::Point xy)
	{
		current_point = xy;
	}
	
	void Velocity::SetThreshold(double new_threshold)
	{
		threshold = new_threshold;
		
		CalculateNewThreshold();
	}
	
	void Velocity::SetMinFixTime(double new_minfixtime)
	{
		minFixTime = new_minfixtime;
	}
	
	void Velocity::SetSampleRate(double new_samplerate)
	{
		sampleRate = new_samplerate;
		
		CalculateNewThreshold();
	}
	
	void Velocity::CalculateNewThreshold()
	{
		// Get threshold in mm/second. Simple trig. We've hard
		// coded the distance from the eye to the screen.
		
		// Note! What is tan(90)? Why, it's undefined! This needs fixed
		// at some point. For right now, we're going to document that the
		// threshold should be between 0 and 90 degrees per second.
		double mmScreen = Velocity::mmEyeToScreen * tan((M_PI * threshold) / 180.0);
		
		// Convert to pixels, since that's how we get eye data.
		velocityThreshold = mmScreen / (mmAcrossScreen / pixelsAcrossScreen);
		
		// Interval changes when frequency changes.
		sampleInterval = 1000.0 / (double)sampleRate;
		
		return;
	}

	/////////////////////////////////////////////////////////////////
	//
	// Returns state of fixation.
	//
	VF::FixationState Velocity::Add(VF::Point xy, bool isEyeFound)
	{
		// First thing. Was the eye found for this sample. If not,
		// then we obviously have to ignore it.
		if (!isEyeFound)
		{
			countIgnoredSamples++;
			return state;
		}
		
		// Find the distance between last good sample and this one.
		double x2 = pow(xy.x - current_point.x, 2);
		double y2 = pow(xy.y - current_point.y, 2);

		// Because our threshold is a velocity, we need to express this in distance/seconds.
		// But, we also need to take into account the number of samples between 
		// the last good sample and this one. In other words, we may have had one 
		// or more 'eye not found' samples so that additional time has gone by.
		// Distance / time -- the 1000.0 is there to get us from milliseconds to seconds.
		double curVelocity = sqrt(x2 + y2) * (1000.0 / (sampleInterval * (countIgnoredSamples + 1)));
		
		// Now look for an "impossible" sample. We may get a true 'eye found flag', but
		// a glitch may make the eye data inaccurate. Assume if we get a velocity of
		// more than 800 degrees/second that it is a sample we should ignore.
		if (curVelocity > impossibleThreshold)
		{
			countIgnoredSamples++;
			return state;
		}

		current_point = xy;	// new point to current point
		
		// Switch state based on velocity.
		if (curVelocity < velocityThreshold)
		{
			// Under the threshold, then add this point to the
			// current fixation list.
			current_fixation.push_back(xy);
			
			switch(state)
			{
				case VF::MOVING:
					{
						// Only switch from moving if enough time has elapsed.
						int		countSamples	= current_fixation.size() + countIgnoredSamples;
						double	elapsedTime		= countSamples * sampleInterval;
						
						if (elapsedTime >= minFixTime)
							state = VF::FIXBEGIN;
					}
					break;
					
				case VF::FIXBEGIN:	// only lasts for one sample
					state = VF::FIX;
					break;
					
				case VF::FIXEND:		// only lasts for one sample
					state = VF::MOVING;
					break;
					
				case VF::FIX:
				default:
					break;
			}
			
		}
		else
		{
			countOutsideFixation++;

			// We're over the threshold. Process a fixation if we had one
			// and switch states accordingly.
			switch(state)
			{
				case VF::FIXBEGIN:
				case VF::FIX:
					if (ProcessEndFixation())
					{
						countOutsideFixation = 0;
						current_fixation.clear();
						state = VF::FIXEND;
					}
					else
						state = VF::FIX;
					
					break;
					
				case VF::FIXEND:
					state = VF::MOVING;
					break;
					
				default:
					break;	// VF::MOVING
			}
		}

		// Reset the count of ignored samples, since we now have a good one.
		countIgnoredSamples = 0;
		
		return state;
	}

	bool Velocity::ProcessEndFixation()
	{
		bool ret = false; // not the end of a fixation
		
		if (countOutsideFixation > maxIgnoreOutsideFixation)
		{
			VF::FixationPoint fp;

			fp.center_gravity = CenterOfGravity(current_fixation);
			fp.num_points		= current_fixation.size();

			fixation_points.push_back(fp);
			ret = true;
		}
		
		return ret;
	}

	VF::Point Velocity::CenterOfGravity(list<VF::Point> &lst)
	{
		VF::Point cg = {0, 0};
		
		long x = 0;
		long y = 0;
		
		list<VF::Point>::iterator iter = lst.begin();
		while (iter != lst.end())
		{
			x += iter->x;
			y += iter->y;
			
			iter++;
		}
		
		cg.x = x / lst.size();
		cg.y = y / lst.size();
		
		return cg;
	}
}
