 ///////////////////////////////////////////////////////////////////
 //
 // File: velocity.h
 // 
 // Project: fixation Max external.
 // 
 // Authors: Erik Brown.
 // Principle Investigator: Anthony Hornof
 // Cognitive Modeling and Eye Tracking Lab
 // CIS Department, University of Oregon
 // 
 // Description: Provides a velocity-based fixation algorithm for
 //				  use in our Max fixation external. Has setters for
 //				  parameters: velocity threshold, minimum fixation
 //				  time, sample rate, so these can be changed
 //				  dynamically.
 //
 // 
 // Copyright (c) 2007 by the University of Oregon.
 // ALL RIGHTS RESERVED.
 //
 // 
 // Permission to use, copy, and distribute this software in
 // its entirety for non-commercial purposes and without fee,
 // is hereby granted, provided that the above copyright notice
 // and this permission notice appear in all copies and their
 // documentation.
 // 
 // Software developers, consultants, or anyone else who wishes
 // to use all or part of the software or its documentation for
 // commercial purposes should contact the Technology Transfer
 // Office at the University of Oregon to arrange a commercial
 // license agreement.
 // 
 // This software is provided "as is" without expressed or
 // implied warranty of any kind.
 // 
 //
 ///////////////////////////////////////////////////////////////////
 
#ifndef VELOCITY_H
#define VELOCITY_H

#include <list>

using namespace std;

// We've got some typdefs and structs. Different platforms will likely
// implement a Point struct, and since we want to be platform independent,
// let's just put our code in a namespace.
namespace VF
{
	// Can be used by the caller if desired. Returned by a call to
	// Velocity::Add().
	typedef enum FixationState
	{
		MOVING, FIXBEGIN, FIX, FIXEND
	} FixationState;

	// Simple point.
	typedef struct Point
	{
		int x;
		int y;
	} Point;

	// A fixation point.
	typedef struct FixationPoint
	{
		Point center_gravity;
		long  num_points;
	} FixationPoint;

	
	///////////////////////////////////////////////
	//
	// Now we've got our class.
	//
	class Velocity
	{
		public:
			Velocity();
			virtual ~Velocity();

			FixationState	Add(Point xy, bool isEyeFound);
			void				SetThreshold(double new_threshold);
			void				SetMinFixTime(double new_minfixtime);
			void           SetSampleRate(double new_samplerate);
			void           SetCurrentPoint(VF::Point xy);
			
			static double  mmEyeToScreen;
			static int     maxIgnoreOutsideFixation;

		protected:
			list<FixationPoint>		fixation_points;
			list<Point>					current_fixation;
			VF::Point					current_point;
			
			VF::Point					CenterOfGravity(list<Point> &lst);
			void							CalculateNewThreshold();
			
			bool							ProcessEndFixation();
			
			double			threshold;
			int				minFixTime;
			double			sampleRate;			// cycle expressed in Hz
			double			sampleInterval;	// interval expressed in milliseconds
			FixationState	state;
			
			// Calculated in the constructor.
			double pixelsAcrossScreen;
			double mmAcrossScreen;

			// Calculated when we get a new threshold or sample rate.
			double velocityThreshold;
			
			// Ignore velocities over this.
			double impossibleThreshold;
			
			// Set this for data outside fixation. Use it to
			// ignore N points outside of an ongoing fixation.
			int	countOutsideFixation;
			
			// Use this for ignoring eye samples. We need to keep track of
			// these because we need to account for the time between good
			// samples in order to get an accurate velocity.
			int	countIgnoredSamples;

		private:
			// Prohibit the use of these until proven necessary.
			Velocity& operator=(const Velocity &source);
			Velocity(const Velocity& source);
	};
}

#endif  // VELOCITY_H

